/*
 * Copyright (C) Tildeslash Ltd. All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 *
 * You must obey the GNU Affero General Public License in all respects
 * for all of the code used other than OpenSSL.
 */

#include "config.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#define safe_strlen(sz)		(sz == NULL ? 0 : strlen(sz))

#include "chk_string_mask.h"
#include "protocol.h"
#include "exceptions/IOException.h"
#include "exceptions/MemoryException.h"
#include "exceptions/ProtocolException.h"


/**
 *  socks5 test. Default port: 1080/tcp
 *  Optional arguments: username password
 *
 *  @file
 */
void check_socks5(Socket_T socket)
{
	Port_T port;
	const char * szUsername	= NULL;
	const char * szPassword	= NULL;

// https://tools.ietf.org/html/rfc1928
	#define REQUEST_LEN	3
	char probe_null[REQUEST_LEN] = "\x05\x01\x00";
	char probe_auth[REQUEST_LEN] = "\x05\x01\x02";
	char * probe;

	#define REPLY_LEN	2
	char reply[REPLY_LEN];

// https://tools.ietf.org/html/rfc1929
	#define LOGIN_MAXLEN		(3 + 255 + 255)
	char requestLogin[LOGIN_MAXLEN];
	int br;

	ASSERT(socket);
	port = Socket_getPort(socket);
	if (port != NULL)
	{
		size_t iLoginLen;

		szUsername = port->parameters.socks5.username;
		szPassword = port->parameters.socks5.password;

		iLoginLen = 3 + safe_strlen(szUsername) + safe_strlen(szPassword);
		if (iLoginLen > LOGIN_MAXLEN)
			THROW(MemoryException, "SOCKS5: internal error (insufficient buffer for request) -- %s", STRERROR);
	}
	if (szUsername == NULL || * szUsername == 0)
		probe = probe_null;
	else
		probe = probe_auth;

	Socket_nodelay(socket);
	if (Socket_write(socket, probe, REQUEST_LEN) <= 0)
		THROW(IOException, "SOCKS5: error sending request -- %s", STRERROR);

	br = Socket_read(socket, reply, REPLY_LEN);
	if (br <= 0)
		THROW(IOException, "SOCKS5: did not receive answer from server -- %s", STRERROR);
	if (br != REPLY_LEN)
		THROW(ProtocolException, "SOCKS5: Received %d bytes from server, expected %d bytes", br, REPLY_LEN);

	if (reply[0] != 5 || reply[1] != probe[2])
		THROW(ProtocolException, "SOCKS5: unknown server response (code %d) -- %s", reply[1], STRERROR);

	if (probe == probe_auth)
	{
		char * ptr;

		requestLogin[0] = 1;
		ptr = &requestLogin[1];
		ptr = strpush_len_szstr(ptr, szUsername);
		ptr = strpush_len_szstr(ptr, szPassword);
		br = ptr - &requestLogin[0];
		if (Socket_write(socket, requestLogin, br) <= 0)
			THROW(IOException, "SOCKS5: error sending login request -- %s", STRERROR);

		br = Socket_read(socket, reply, REPLY_LEN);
		if (br <= 0)
			THROW(IOException, "SOCKS5: did not receive answer from server -- %s", STRERROR);
		if (br != REPLY_LEN)
			THROW(ProtocolException, "SOCKS5: received %d bytes from server, expected %d bytes", br, REPLY_LEN);

		if (reply[0] != 1 || reply[1] != 0)
			THROW(ProtocolException, "SOCKS5: unknown server response to login (code %d) -- %s", reply[1], STRERROR);
	}
}
